{-

SPS/APS - Security Protocol Specification langauge
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2013-2015

All Rights Reserved.

-}
module ToPv where
import Analyze
import ComposeM
import StrandAST
import Msg
import Ast
import Preprocessor
import GeneratePlainStrands
import OperationalStrands
import Data.List
import Data.Char
import Data.List.Split
------------------------------------------------------------------------------------------- 
toPIsFinal conds ag strnds = theHeader (map rest strnds) strnds ++ 
                                                        toPis conds strnds strnds ++ 
                                                        theFooter ag (owners strnds) (map (fltrFrsh (owners strnds)) (knowledges strnds))
-- show( rmvdups ( (concat $ map getSignatures (terms (concat ( knowledges strands)))) ++ concat(concat (getStrndsSigns (map rest strands)))))
--------------------------- THE HEADER -----------------------------------------------------
theHeader x strands = disclChanIntr ++ funDecProp  True (getSignsActsAndKnows x strands) ++ channels ++ decSecrets(rmvdups(allSecretIDs x)) ++ (decOfAllEvents x)
-------------------------------------------------------------------------------------------
disclChanIntr = "(*Auto-Generated by APS2Java*)\nfree ch: channel.\nfree i:bitstring.\n" --"param traceDisplay = long.\nfree c. \n"
-------------------------------------------------------------------------------------------
getSignsActsAndKnows x ss= rmvdups ( (concat $ map getSignatures (terms (concat ( knowledges ss)))) ++ concat(concat (getStrndsSigns x)))
-------------------------------------------------------------------------------------------
terms k = [b| (a,b)<-k]
-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
decSecrets [] ="\n"
decSecrets (x:xs)= ("free " ++ x ++":bitstring[private].\n") ++ decSecrets xs ++ ("query attacker(" ++ x ++ ").\n")
-------------------------------------------------------------------------------------------
allSecretIDs  (x:xs)=secretIDs x ++ allSecretIDs xs
allSecretIDs  []=[]
-------------------------------------------------------------------------------------------
secretIDs (Add Secrt id m p rs)        = (idID id) : (secretIDs rs)
secretIDs (Add e id m ps rs)        = (secretIDs rs)
secretIDs (Fresh tps ms orms rs)                = (secretIDs rs)
secretIDs (EOS _)                                = []
secretIDs (Snd _ _ rs)                         = (secretIDs rs)
secretIDs (Rcv _ _ rs)                         = (secretIDs rs)
secretIDs (V _ _ _ rs)                        = (secretIDs rs)
secretIDs (Create _ _ _ rs)         = (secretIDs rs)
secretIDs (How _ _ _ _ rs)                 = (secretIDs rs)
-------------------------------------------------------------------------------------------
-- chk list is to eliminate duplicated secret declaration
decOfAllEvents (x:xs)=decOfEvents x ++ decOfAllEvents  xs
decOfAllEvents  []="\n"
-------------------------------------------------------------------------------------------
decOfEvents (Add Secrt id m p rs)    = decOfEvents rs
decOfEvents (Add Witness id m ps rs)    = "event witness" ++(idID id) ++"("++ (bitstrs ((length ps)+1)) ++ ").\n" ++ (decOfEvents rs) 
decOfEvents (Add Request id m ps rs)= "event request" ++(idID id) ++"("++ (bitstrs ((length ps)+1)) ++ ").\n" ++ (decOfEvents rs) 
                                                                                ++ "query m: bitstring, p1: bitstring, p2: bitstring; inj-event(request"++(idID id)++"(m,p1,p2)) ==> inj-event(witness"++ (idID id) ++"(m,p1,p2)).\n"
decOfEvents (Add Wrequest id m ps rs)= "event wrequest" ++(idID id) ++"("++ (bitstrs ((length ps)+1)) ++ ").\n" ++ (decOfEvents rs) 
                                                                                ++ "query m: bitstring, p1: bitstring, p2: bitstring; event(wrequest"++(idID id)++"(m,p1,p2)) ==> event(witness"++ (idID id) ++"(m,p1,p2)).\n"
decOfEvents (Fresh tps ms orms rs)                = decOfEvents rs
decOfEvents (EOS _)                                        = ""
decOfEvents (Snd _ _ rs)                         = decOfEvents rs
decOfEvents (Rcv _ _ rs)                         = decOfEvents rs
decOfEvents (V _ _ _ rs)                        = decOfEvents rs
decOfEvents (Create _ _ _ rs)                 = decOfEvents rs
decOfEvents (How _ _ _ _ rs)                 = decOfEvents rs

-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
-- funDecProp functions declarations and properties
-- funDecProp: takes a flag (to insure that inv proceeds crypt and sign) and a list of operators 
-- the list of operators includes mappings, crypto API, formats, mult and exp 
-- and lists their properites which functions to decalre and which to leave out!
funDecProp f (("inv",arity):xs)                 = if(f==True) then invProp ++ funDecProp False xs else funDecProp f xs
funDecProp f (("pk",arity):xs)                         = pkProp          ++ funDecProp f xs
funDecProp f (("shk",arity):xs)                 = shkProp          ++ funDecProp f xs
funDecProp f (("sk",arity):xs)                         = skProp          ++ funDecProp f xs                  
funDecProp f (('m':'p':'p':mp,arity):xs)= "fun "         ++ lowerize mp ++ "("++ bitstrs arity ++ "):bitstring[private].\n" ++ funDecProp f xs -- end of Mappings
funDecProp f (("crypt",arity):xs)                 = if(f==True) then invProp ++ cryptProp ++ funDecProp False xs else  cryptProp ++ funDecProp f xs
funDecProp f (("scrypt",arity):xs)                 = funDecProp f xs         --we always need is for secrecy, so added at the end
funDecProp f (("sign",arity):xs)                 = if(f==True) then invProp ++ signProp ++ funDecProp False xs else signProp ++ funDecProp f xs
funDecProp f (("mac",arity):xs)                 = macProp         ++ funDecProp f xs
funDecProp f (("hash",arity):xs)                 = hashProp  ++ funDecProp f xs                -- end of crypto API
funDecProp f (("exp",arity):xs)                 = expProp          ++ funDecProp f xs
funDecProp f (("mult",arity):xs)                 = multProp  ++ funDecProp f xs                -- end of algebaric operators
funDecProp f ((('f':'r':'m':fun),ar):xs)= "fun "         ++ lowerize fun ++ "("++ bitstrs ar ++ "):bitstring.\n"  ++ getters (lowerize fun)  1 ar ++ funDecProp f xs -- formats
funDecProp f ((fun,arity):xs)                         = "fun "         ++ lowerize fun  ++ "("++ bitstrs arity ++ "):bitstring.\n" ++ funDecProp f xs
funDecProp f []                                                 = scryptProp++"\n(* END OF FUNCSansPROPS*)\n\n" 
-------------------------------------------------------------------------------------------
-- getters: takes a format name, in integer counter (starts at 1) and the arity, and produces the getters properties
getters fun counter total= if(counter>total) then "" else   --"fun "++fun++"get"++show(counter)++"(bitstring):bitstring.\n\t"++
                                        "\treduc forall " ++(forAll (take total alphabetNog))++fun++"get"++show(counter)++ "("++fun++"("++ (lst(take total alphabetNog))++"))=" 
                                        ++ toStr(alphabetNog!!(counter-1)) ++ ".\n"
                                        ++getters fun (counter +1) total
-------------------------------------------------------------------------------------------
bitstrs n= if(n==1) then "bitstring" else "bitstring," ++ bitstrs (n-1)
-------------------------------------------------------------------------------------------
alphabet = ['a'..'z']
alphabetNog = ['a'..'z']\\['g','i'] -- Special alphabet to execlude g from the getters parameteres!
------------------------------------------------------------------------------------------- 'A'
lst s@(x:[])= s
lst s@(x:xs)= [x]++ "," ++ lst xs
-------------------------------------------------------------------------------------------
forAll s@(x:[])= s ++":bitstring;"
forAll s@(x:xs)= [x]++ ":bitstring," ++ forAll xs
-------------------------------------------------------------------------------------------
forAlll [] = ""
forAlll (x:[])= lowerize x ++":bitstring"
forAlll (x:y:xs)= lowerize x ++ ":bitstring," ++ forAlll (y:xs)
-------------------------------------------------------------------------------------------
invProp                =         "fun inv(bitstring):bitstring [private].\n"                         -- inv
pkProp                =         "fun pk(bitstring):bitstring[private].\n"                                 -- pk as public
shkProp                =         "fun shk(bitstring,bitstring):bitstring [private].\n"         -- shk
skProp                =         "fun sk(bitstring):bitstring [private].\n"                                 -- sk
-- pbkProp                =         "fun pbk(bitstring):bitstring.\n"                                         -- pbk
-- pvkProp                =         "fun pvk(bitstring):bitstring.\n"                                         -- pvk
-- pubProp                =         "fun pub(bitstring):bitstring.\n"                                         -- pub
cryptProp        =        "fun crypt(bitstring,bitstring):bitstring. \n\t" ++
                                "reduc forall k: bitstring, m: bitstring; decrypt(inv(k),crypt(k,m))=m.\n"                -- crypt
scryptProp        =         "fun scrypt(bitstring,bitstring):bitstring. \n\t"++
                                        "reduc forall k: bitstring, m: bitstring;  sdecrypt(k,scrypt(k,m))=m.\n"        -- scrypt
signProp        =         "fun sign(bitstring,bitstring):bitstring. \n\t" ++
                                        "reduc forall k: bitstring, m: bitstring; open(k,sign(inv(k),m))=m.\n"                -- sign
macProp         =        "fun mac(bitstring,bitstring):bitstring. \n"                                                                         -- mac
hashProp        =        "fun hash(bitstring):bitstring. \n"                                                                                                -- hash
expProp                =        "const g: bitstring.\n" ++
                                        "fun exp(bitstring,bitstring):bitstring.\n\tequation "++
                                        "forall  x: bitstring, y: bitstring; exp(exp(g,x),y)=exp(exp(g,y),x).\n"        -- exp with fixed g
multProp        =        "const g:bitstring.\nfun mult(bitstring,bitstring):bitstring. \n\t"++
                                        "equation forall  a: bitstring, b: bitstring; mult(a,b)=mult(b,a).\n"   
										-- mult1                        
						++ "equation forall  a: bitstring, b: bitstring, c:bitstring; mult(a, mult(b,c)) =mult(mult(a,b),c).\n"                 -- mult2
channels        =         "fun secCh(bitstring,bitstring): channel[private].\n" ++
                                "fun confCh(bitstring): channel[private].\n" ++ -- receiver
                                "fun authCh(bitstring): channel.\n" --sender Not private!
-------------------------------------------------------------------------------------------        
toStr a=[a]
-- Just a footer
-- theFooter x = "process " ++ aFooter x x
theFooter ag ownrs knows = "process\n"++(decConsts ag knows)++"\n!new x:bitstring;\nout(ch, x)|\n" ++ instAgents ownrs ownrs knows 
-------------------------------------------------------------------------------------------
-- instAgents x [a] [k] = """!in(ch,(" ++(forAll ) ++");
instAgents x (o1:[]) (k1:rst) = if(isNotVar o1) then inCh ++("process"++o1++ "("++(gtIdntPvs (map snd k1)) ++")\n") 
                                                                 else inCh ++("process"++o1++ "("++(gtIdntPvs procParams) ++") | \nout(ch,("++ (gtIdntPvs (rplcMcros (iLet o1) (map snd k1))) ++"))\n")
                                                                        where procParams=(rplcMcros (xLet o1) (map snd k1));
                                                                                        -- nonOwners= x\\[o1];        usedRoles= onlyRoles `intersect` procParams ;
                                                                                                nonOwners= x\\[o1];
                                                                                                onlyRoles= (filter isVar nonOwners)\\[o1];
                                                                                                usedRoles= onlyRoles `intersect` (getAtomic procParams) ;                                                                                        
                                                                                        inCh= if(usedRoles==[]) then "!" else "!in(ch,("++ (forAlll usedRoles)++"));\n";        
-- (("!in(ch,("++ (forAlll (x\\[o1]))++"))("!in(ch,("++ (forAlll (x\\[o1]))++"));rmvdups $ getAtomic
instAgents x (o1:o2:orst) (k1:k2:krst)= if(isNotVar o1) then inCh ++("process"++o1++ "("++(gtIdntPvs (map snd k1)) ++")\n") 
                                                                                        else inCh++"process"++o1++ "("++(gtIdntPvs procParams)++") | \nout(ch,("++ (gtIdntPvs (rplcMcros (iLet o1) (map snd k1))) ++"))|\n"++(instAgents x (o2:orst) (k2:krst))
                                                                                where procParams= (rplcMcros (xLet o1) (map snd k1));
                                                                                                nonOwners= x\\[o1];
                                                                                                onlyRoles= (filter isVar nonOwners)\\[o1];
                                                                                                usedRoles= onlyRoles `intersect` (getAtomic procParams) ;
                                                                                                inCh= if(usedRoles==[]) then "!" else "!in(ch,("++ (forAlll usedRoles)++"));\n";
-------------------------------------------------------------------------------------------
-- takes a list of owners and knowleldges, extracts constants non-agents  and declare them in the form c:bitstring;
-- decConsts os ks = "\n\nHello" ++(concat (getConsts ks)) ++"\n\nHello"
decConsts os ks =newValues' $ rmvdups ((filter isNotVar os) ++ (concat (getConsts ks))) 
-------------------------------------------------------------------------------------------
getConsts []=[]
getConsts (k1:rest) = ((getAtomIdents (map snd k1))) : getConsts rest
-------------------------------------------------------------------------------------------
getAtomIdents [] =[]
getAtomIdents ((Atom x):xs)= if(isVar x) then getAtomIdents xs else (x:getAtomIdents xs)
getAtomIdents ((Comp op t lst):xs)= getAtomIdents (xs++lst)
-- Comp Operator TypeOpts [Msg]
-------------------------------------------------------------------------------------------
------------------------------ THE PROCESSES ----------------------------------------------
-------------------------------------------------------------------------------------------
toPis conds [] s= "\n" --        ++"\n\nCONDITIONS" ++ show(conds) ++ "END \n"
toPis conds s@(x:xs) ss=   (toPi x (owners ss)) ++ (toPis conds xs ss) 
-------------------------------------------------------------------------------------------                
isaVar (Atom x)= isVariable x
isaVar _ =False
-- toPi:: [Types] -> Strand -> String -- toPi::  Strand -> String
toPi s@(Strand n o know rs) ownrs= (header n o know ownrs) ++ (toPi0 (length know) (getLbls (ownrs\\[o]) know) [] rs) 
-------------------------------------------------------------------------------------------
-- toPi translates strands to pv code. chks is a list of checks in order to eliminate duplicated check gtIdntPvs 
toPi0 n g chks (Snd det m rs)       = "out("++ chDet det++", " ++ lowerize (gtIdntPv m)++ ");\n"++ (toPi0 n g chks rs)
toPi0 n g chks (Rcv det m rs)        =  "in("++ chDet det++", " ++ lowerize (gtIdntPv m) ++ ":bitstring);\n" ++ (toPi0 n g chks rs)
toPi0 n g chks (Fresh tps ms orms rs)    = newValues ms ++ invPubs tps n ms ++ (toPi0 n g chks rs) 
toPi0 n g chks s@(V Check m1 m2 rs)     =  if (elem ("Check",m1,m2) chks || elem ("Check",m2,m1) chks) 
                                                                                        then toPi0 n g chks rs 
                                                                                else "if(" ++ lowerize (gtIdntPv m1)++ " = " ++ lowerize(gtIdntPv m2) ++ ") then\n" ++ 
                                                                                        (toPi0 n g (("Check",m1,m2):chks) rs)
toPi0 n g chks s@(V ver m1 m2 rs)         =  toPi0 n g chks  rs
toPi0 n g chks (Create m1 m2 m3 rs)        =  (toPi0 n g chks rs)
toPi0 n g chks (How (Get nn) m1 m2 m3 rs)= "let "++ lowerize (gtIdntPv m1)++":bitstring = "++lowerize(gtIdntPv m2)++"get" ++ show(nn)
                                                                                        ++ "("  ++ lowerize (gtIdntPv m3) ++ ") in \n" ++ (toPi0 n g chks rs)
toPi0 n g chks (How (Project nn) m1 m2 m3 rs) = "let "++ lowerize (gtIdntPv m1) ++ ":bitstring = proj" ++ show(nn)++ "("  ++ lowerize (gtIdntPv m2) ++                                                                                                         ") in \n" ++ (toPi0 n g chks rs)
toPi0 n g chks (How des m1 m2 m3 rs)=         "let "++ lowerize (gtIdntPv m1) ++ ":bitstring = " ++ lowerize(show(des)) ++"(" ++ lowerize (gtIdntPv m2)++ ", " ++ 
                                                                                                lowerize (gtIdntPv m3)++") in \n" ++ (toPi0 n g (("Open",m2,m3):chks) rs)
-- toPi0 g chks (How des m1 m2 m3 rs)=         "let "++ lowerize (gtIdntPv m1) ++ ":bitstring = " ++ lowerize (show(des))++"(" ++ lowerize (gtIdntPv m2)++ ", " ++ 
                                                                                                -- lowerize (gtIdntPv m3)++") in \n" ++ (toPi0 g (("Open",m2,m3):chks) rs)
toPi0 n g chks (Add Witness id m ps rs)= "event witness" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPi0 n g chks rs)
toPi0 n g chks (Add Secrt id m ps rs)= if (guards g =="") then b else a ++ b
                                                                                where a= "if("++(guards g)++") then \n";
                                                                                        b= "out(ch, scrypt("++lowerize (gtIdntPv m)++","++(idID id)++"));" ++ (toPiRest rs)
toPi0 n g chks (Add Request id m ps rs)= if (guards g =="") then b else a ++ b
                                                                                where a="if("++(guards g)++") then \n";
                                                                                        b="event request" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPiRest rs);
toPi0 n g chks (Add Wrequest id m ps rs)=  if (guards g =="") then b else a ++ b
                                                                                where a="if("++(guards g)++") then \n";
                                                                                        b="event wrequest" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPiRest rs);
-- "if("++(guards g)++") then \nevent wrequest" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPiRest rs)
toPi0 n g chks (EOS _)                                 =  "0.\n(* -------------------------- *)\n\t\t(*EOP*)\n" --(* ------------------ *)\n\n"
-------------------------------------------------------------------------------------------
toPiRest (Add Witness id m ps rs)= "event witness" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPiRest rs)
toPiRest (Add Secrt id m ps rs)= "out(ch, scrypt("++lowerize (gtIdntPv m)++","++(idID id)++"));" ++ (toPiRest rs)
toPiRest (Add Request id m ps rs)= "event request" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPiRest rs)
toPiRest (Add Wrequest id m ps rs)= "event wrequest" ++ (idID id) ++ "("++ lowerize (gtIdntPv m)++"," ++ (gtIdntPvs ps)++ ");\n"++(toPiRest rs)
toPiRest (EOS _)                                 =  "0.\n(* -------------------------- *)\n\t\t(*EOP*)\n" --(* ------------------ *)\n\n"
toPiRest  s                                 =  error("There is a problem, event followed by exchange!\n"++show(s))
-------------------------------------------------------------------------------------------
guards [] = error("WHAT IS GOING ON HERE??!!\n")
guards ("":g2:rst)= guards (g2:rst)
guards [g]= if(g=="") then "" else g++"<> i"
guards (g1:"":rst)= guards (g1:rst)
guards (g1:g2:rst)= g1++ "<> i &&" ++ guards (g2:rst)
-------------------------
-- here we create a private key for fresh public keys
invPubs (PublicKey _) n [x] = "let x"++ show(n+2) ++" = inv("++ (lowerize (gtIdntPv x)) ++") in\n" 
invPubs (PublicKey p) n (x:xs) ="let x"++ show(n+2) ++" = inv("++ (lowerize (gtIdntPv x)) ++") in\n" ++ (invPubs (PublicKey p) (n+1) (xs)) 
invPubs _ n _ = ""
-------------------------
-------------------------
chDet ((sender,Secure,receiver), _,_, _) = "secCh("++ nameOfPeer sender++","++ nameOfPeer receiver++")" 
chDet ((sender,Authentic,receiver), _,_, _) = "authCh("++ nameOfPeer sender++")"
chDet ((sender,Confidential,receiver), _,_, _) = "confCh("++ nameOfPeer receiver++")"
chDet ((_,_,_), _,_, _) = "ch"
----------------------------------
-------------------------------------------------------------------------------------------
nameOfPeer (_,a,_,_)=lowerize $ getIdent a
-------------------------------------------------------------------------------------------
idID id= cleanString ((id!!0)++(id!!2) ++ (id!!3))
-------------------------------------------------------------------------------------------
cleanString st = [ c | c <- st, c /= ',' && c/= ')' && c/='(' && c/=' ']
-------------------------------------------------------------------------------------------
header n o k owners= "(*Agent "++ o ++" process in protocol:"  ++ n ++"  is *)\nlet process" ++ o ++"("++ (params (fltrFrsh owners k)) ++") =\n" 
-------------------------------------------------------------------------------------------
params [(a,b)] = lowerize (gtIdntPv a) ++ ":bitstring"
params ((a,b):y:z) = lowerize (gtIdntPv a)++ ":bitstring," ++ (params(y:z))
-------------------------------------------------------------------------------------------
ppDes des m1 m2 m3= "\n\nHHH"++ show(m1)++show(m2)++show(m3)++"HHH\n\n"
-- ttsOOS
-------------------------------------------------------------------------------------------
-- callparams: takes a knowlede, and return its terms as a comma-separated list
callparams [(a,b)] = lowerize (gtIdntPv b) 
callparams ((a,b):y:z) = lowerize (gtIdntPv b) ++"," ++ (callparams(y:z))
-------------------------------------------------------------------------------------------
-- rplcparamsI: takes a special macro 'x' and replace all occurences of [a] with 'x'
rplcparamsX  xlet know = rplcMcros xlet know
xLet a = ("WereLet", [a], (Atom "X"))
-------------------------------------------------------------------------------------------
-- rplcparamsI: takes a special macro 'i' and replace all occurences of [a] with 'i'
rplcparamsI iLet know = rplcMcros iLet know
-------------------------------------------------------------------------------------------
-- fltrFrsh: takes a list of owners and a knowledge, and returns only roles
fltrFrsh owners k  = [(a,b)| (a,b)<-k,  isOwner b owners || not (isaVar b)]
-------------------------------------------------------------------------------------------
-- isOwner: takes a message and and a list of idents (owners), and checks if the message is tha name on an owner!
isOwner (Atom x) o= elem x o
isOwner m o= False
-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
-- getLbls: takes a list of messages and a knowledge, and returns a list of their labels in that given knowledge
getLbls [] k=[]
getLbls (m:sgs) k= (getLbl m k) : (getLbls sgs k)
-- getLbls msgs k = map (getLbl msgs) k
-------------------------------------------------------------------------------------------
-- getLbl: takes a message and a knowledge, and returns its label in that given knowledge
getLbl m k= concat [lowerize (gtIdntPv a)| (a,b)<-k, b==Atom m] 
-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
-- getIdents' : takes a list of format names and returns a list of their original names
-- for formats (special case), we appended an 'a' at the end of each variable to contain a format, so this getIdents' takes off that 'a'
getIdents' []=""
getIdents' [x]= lowerize (getIdent x)
getIdents' (x:y:z)= lowerize (getIdent x) ++ "_" ++ (getIdents'(y:z))
-------------------------------------------------------------------------------------------
-- getIdent': (special case of getIdent for formats) takes a format name and returns its original name (taking-off the appended 'a')
-- the next one is special for received formats labels, i.e., to take-off the appended 'a'
getIdent' (Atom x)= init x
-------------------------------------------------------------------------------------------
-- newValues': takes a list of idents and returns a string listing of new declarations
-- takes a list of fresh values and produces a list of them
newValues' [] =""
newValues' (x:xs)= if(x=="g") then newValues' xs else "new " ++ lowerize x ++":bitstring;\n" ++ newValues' xs -- removing g the base constant of exp!!
-------------------------------------------------------------------------------------------
-- newValues: takes a list of messages and returns a string listing of new declarations
newValues [] ="\n"
newValues (x:xs)= "new " ++ lowerize (gtIdntPv x) ++":bitstring;\n" ++ newValues xs
-------------------------------------------------------------------------------------------
-- For a list of messages, getSignatures gives the operations in there!
getSignatures (Comp Apply opt ((Atom x):xs))= (x,length xs): concat(map getSignatures xs)
getSignatures (Comp Cat opt ls)= concat(map getSignatures ls) 
getSignatures (Atom x)= [] --Nothing --"(" ++ (getSignaturess ls)++")"
getSignatures (Comp Hash opt xs)= ("hash",1): concat (map getSignatures xs)
getSignatures (Comp Mac opt xs)= ("mac",2): concat (map getSignatures xs)
getSignatures (Comp op opt xs)= (lowerize $ show(op),length xs): concat (map getSignatures xs)
-------------------------------------------------------------------------------------------
-- getAllSignatures: takes a strand, and produces all signatures in it
getAllSignatures  (Snd det m rs)                                         = getSignatures m : (getAllSignatures  rs)
getAllSignatures  (Rcv det m rs)                                         = getSignatures m :  (getAllSignatures  rs)
getAllSignatures  (Fresh tps ms orms rs)                         = (getAllSignatures  rs)
getAllSignatures  (V Check m1 m2 rs)                                 =  (getAllSignatures rs)
getAllSignatures  (V ver m1 m2 rs)                                         =  getAllSignatures   rs
getAllSignatures  (Create m1 m2 m3 rs)                                = getSignatures m3 :(getAllSignatures rs)
getAllSignatures  (How (Get n) m1 m2 m3 rs)                 = (getAllSignatures rs)
getAllSignatures  (How (Project n) m1 m2 m3 rs)         = (getAllSignatures rs)
getAllSignatures  (How des m1 m2 m3 rs)                                =  (getAllSignatures rs)
getAllSignatures  (Add event id m ps rs)                        = (getAllSignatures rs)
getAllSignatures  (EOS _)                                                         =  []
-------------------------------------------------------------------------------------------
-- getStrndsSigns: takes a list of strands and produces all signatures in them
getStrndsSigns []=[]
getStrndsSigns (x:xs)= getAllSignatures x : getStrndsSigns xs
-------------------------------------------------------------------------------------------
-- lsst: takes a list of strings and produces a comma separated list of them
lsst [[x]]=[x]
lsst ([x]:[y]:z)= [x] ++ ", " ++ lsst ([y]:z)
-------------------------------------------------------------------------------------------
-- getAtomic: for a list of msgs, it produces only atomic parts 
getAtomic []=[]
getAtomic ((Atom x):xs) = x : (getAtomic xs)
getAtomic ((Comp op opt ls):xs)= (getAtomic ls) ++ getAtomic xs 
-------------------------------------------------------------------------------------------
gtIdntPv (Atom ('f':'r':'m':rst))= lowerize rst
gtIdntPv (Atom ('m':'p':'p':rst))= lowerize rst
gtIdntPv (Atom x)= lowerize x
gtIdntPv (Comp Apply opt ((Atom ('f':'r':'m':rst)):xs))= lowerize rst++"(" ++(gtIdntPvs xs)++")"
gtIdntPv (Comp Apply opt (x:xs))= (gtIdntPv x) ++ "(" ++ (gtIdntPvs xs)++")"
gtIdntPv (Comp Cat opt ls)= "(" ++ (gtIdntPvs ls)++")"
gtIdntPv (Comp Mac opt ls)= "mac(" ++ gtIdntPv (head ls)++",(" ++ (gtIdntPvs (tail ls))++"))"
gtIdntPv (Comp Hash opt ls)= "hash((" ++ (gtIdntPvs ls)++"))"
gtIdntPv (Comp op opt ls)= lowerize (show(op)) ++ "(" ++ (gtIdntPvs ls)++")"
----------------------
gtIdntPvs []=" "
gtIdntPvs [x]= lowerize (gtIdntPv x)
gtIdntPvs (x:y:z)= lowerize (gtIdntPv x) ++ ", " ++ (gtIdntPvs(y:z))
-------------------------------------------------------------------------------------------