{-

  APCC  

  Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015 
  Minor Updates by Omar Almousa 2015

-}

module TypeSystem_Evaluator (typeof_ts) where

import TypeSystem_Defs
import TypeSystem_Context
import Debug.Trace()
import Msg
import StrandAST

typeError :: Msg -> TermType -> TermType -> Term -> Context -> String
typeError a exptype ty term ctx = show a ++ " must be of type " ++ show exptype ++ "\n" 
                                        ++ "type: " ++ show ty ++ "\n"
                                        ++ "term: " ++ show term ++ "\n"
                                        ++ "ctx: " ++ show ctx    

-- Typing valuation

typeof_ts :: Term -> Context -> TermType
--typeof_ts term ctx | trace ("typeof_ts\n\tterm: " ++ show term ++ "\n\tcontext:" ++ show ctx) False = undefined 

typeof_ts (TmVar id (VarBind _)) ctx  = let (_,VarBind ty) = getBindingByName ctx id in ty

-- messages
typeof_ts (AMsg (Atom ('f':'r':'m':m))) ctx = let (_,VarBind ty) = getBindingByName ctx m in ty
typeof_ts (AMsg (Atom ('m':'p':'p':m))) ctx = let (_,VarBind ty) = getBindingByName ctx m in
                                                        case ty of
                                                            TMapping _ (_,t) -> t
                                                            _ -> error (show ty ++ " must be a Mapping")
typeof_ts (AMsg (Atom m)) ctx = let (_,VarBind ty) = getBindingByName ctx m in ty
                               
typeof_ts (AMsg (Comp Hash _ [a])) ctx = THash (typeof_ts (AMsg a) ctx)
typeof_ts (AMsg (Comp Mac _ [a,b])) ctx = TMac (typeof_ts (AMsg a) ctx, typeof_ts (AMsg b) ctx)
typeof_ts (AMsg (Comp Mult _ [a,b])) ctx = TMult (typeof_ts (AMsg a) ctx, typeof_ts (AMsg b) ctx)
typeof_ts (AMsg (Comp Exp _ [a,b])) ctx = TExp (typeof_ts (AMsg a) ctx, typeof_ts (AMsg b) ctx)
typeof_ts (AMsg (Comp Inv _ [a])) ctx = case typeof_ts (AMsg a) ctx of
                                                TPublicKey -> TPrivateKey
                                                _ -> error (show a ++ " must be a PublicKey")
typeof_ts t@(AMsg (Comp Pk _ [a])) ctx = case typeof_ts (AMsg a) ctx of
                                                TAgent -> TPublicKey
                                                ty -> error (typeError a TAgent ty t ctx) 


typeof_ts t@(AMsg (Comp Sk _ [a])) ctx = case typeof_ts (AMsg a) ctx of
                                                TAgent -> TSymmetricKey
                                                ty -> error (typeError a TAgent ty t ctx)    

typeof_ts t@(AMsg (Comp Shk _ [a,b])) ctx = case typeof_ts (AMsg a) ctx of
                                                TAgent -> case typeof_ts (AMsg b) ctx of
                                                            TAgent -> TSymmetricKey
                                                            tyb -> error (typeError b TAgent tyb t ctx) 
                                                tya -> error (typeError a TAgent tya t ctx) 

typeof_ts (AMsg (Comp Scrypt _ [a,b])) ctx = TScrypt (typeof_ts (AMsg a) ctx,typeof_ts (AMsg b) ctx)


typeof_ts t@(AMsg (Comp Crypt _ [a,b])) ctx = case typeof_ts (AMsg a) ctx of
                                                TPublicKey -> TCrypt (TPublicKey,typeof_ts (AMsg b) ctx)
                                                ty -> error (typeError a TPublicKey ty t ctx)
                                                     
typeof_ts t@(AMsg (Comp Pub _ [a])) ctx = case typeof_ts (AMsg a) ctx of
                                                TPrivateKey -> TPublicKey
                                                ty -> error (typeError a TPrivateKey ty t ctx)   
                                                
typeof_ts t@(AMsg (Comp Sign _ [a,b])) ctx = case typeof_ts (AMsg a) ctx of
                                                TPrivateKey -> TSign (TPrivateKey,typeof_ts (AMsg b) ctx)
                                                ty -> error (typeError a TPrivateKey ty t ctx)   

typeof_ts (AMsg (Comp Cat _ xs)) ctx = TCat (map (\x -> typeof_ts (AMsg x) ctx ) xs)    -- actually Cat is not allowed in APS

typeof_ts t@(AMsg (Comp SignH _ [a,b])) ctx = case typeof_ts (AMsg a) ctx of
                                                TPrivateKey -> case typeof_ts (AMsg b) ctx of
                                                                        t@(THash _) -> TSign (TPrivateKey,t)
                                                                        _ -> error (show b ++ " must be a Hash")
                                                ty -> error (typeError a TPrivateKey ty t ctx)   
typeof_ts t@(AMsg (Comp FsignH _ [a,b])) ctx = case typeof_ts (AMsg a) ctx of
                                                TPrivateKey -> case typeof_ts (AMsg b) ctx of
                                                                        t@(TSign (TPrivateKey,THash _)) -> TSign (TPrivateKey,t)
                                                                        _ -> error (show b ++ " must be a Signature of Hash")
                                                ty -> error (typeError a TPrivateKey ty t ctx)   


typeof_ts t@(AMsg (Comp Apply _ (a@(Atom ('m':'p':'p':rst)):xs))) ctx = case typeof_ts (AMsg (Atom rst)) ctx of
                                                                                TMapping _ (_,t) -> t
                                                                                ty -> error ("mapping " ++ show rst ++ " does not exist or has the wrong type" ++ "\n" 
                                                                                    ++ "type: " ++ show ty ++ "\n"
                                                                                    ++ "term: " ++ show t ++ "\n"
                                                                                    ++ "ctx: " ++ show ctx)     
typeof_ts t@(AMsg (Comp Apply _ ((Atom ('f':'r':'m':rst)):xs))) ctx = case typeof_ts (AMsg (Atom rst)) ctx of
                                                                        TFormat _ _ -> TFormat ident types 
                                                                                          where 
                                                                                            ident = rst
                                                                                            types = map (\x -> typeof_ts (AMsg x) ctx ) xs        
                                                                        ty -> error ("format " ++ show rst ++ " does not exist or has the wrong type" ++ "\n" 
                                                                                    ++ "type: " ++ show ty ++ "\n"
                                                                                    ++ "term: " ++ show t ++ "\n"
                                                                                    ++ "ctx: " ++ show ctx)       
typeof_ts t@(AMsg (Comp Apply _ (Atom x : xs))) ctx = case typeof_ts (AMsg (Atom x)) ctx of
                                                                        TFunction _ _ -> TFunction ident types
                                                                                          where 
                                                                                            ident = x
                                                                                            types = map (\x -> typeof_ts (AMsg x) ctx ) xs
                                                                        -- t@(TMapping _ _) -> t                                                                                                
--                                                                          TFormat _ _ -> TFormat ident types 
--                                                                                          where 
--                                                                                            ident = x
--                                                                                            types = map (\x -> typeof_ts (AMsg x) ctx ) xs                            
                                                                        ty -> error ("function1 " ++ show x ++ " does not exist or has the wrong type" ++ "\n" 
                                                                                    ++ "type: " ++ show ty ++ "\n"
                                                                                    ++ "term: " ++ show t ++ "\n"
                                                                                    ++ "ctx: " ++ show ctx)

typeof_ts t@(AMsg (Comp op _ _)) _ = error ("type of " ++ show t ++ " is undefined\n" 
                                                       ++ "op: " ++ show op  ++ " unhandled")
-- destructors

typeof_ts (ADes Open m1 m2) ctx = case typeof_ts (AMsg m1) ctx of
                                            TPublicKey -> case typeof_ts (AMsg m2) ctx of
                                                            TSign (TPrivateKey,t) -> t
                                                            _ -> error (show m2 ++ " must be a Sign(PrivateKey,t)") 
                                            _ -> error (show m1 ++ " must be a PublicKey")
typeof_ts (ADes Decrypt m1 m2) ctx = case typeof_ts (AMsg m1) ctx of
                                                        TPrivateKey -> case typeof_ts (AMsg m2) ctx of
                                                                        TCrypt (TPublicKey,t) -> t
                                                                        _ -> error (show m2 ++ " must be a Crypt(PublicKey,_)") 
                                                        _ -> error (show m1 ++ " must be a PrivateKey")
typeof_ts (ADes Sdecrypt m1 m2) ctx = let 
                                            t1 = typeof_ts (AMsg m1) ctx
                                            t2 = typeof_ts (AMsg m2) ctx 
                                      in case t2 of
                                                TScrypt (t3,t4) -> if compareTypes t1 t3 then t4 else error (show m2 ++ " must be a Scrypt(" ++ show t1 ++ ",_)") 
                                                _ -> error (show m2 ++ " must be a Scrypt(" ++ show t1 ++ ",_)")  

typeof_ts t@(ADes (Get int) (Atom ('f':'r':'m':id)) m2) ctx = case typeof_ts (AMsg m2) ctx of
                                                                TFormat idf xs -> if idf == id && length xs >= int then xs !! (int - 1) else
                                                                                          error ("index too large/lenght: " ++ show int ++ "/" ++ show (length xs) ++  
                                                                                                  "or wrong format " ++ show id ++ "/" ++ show idf ++ "\n" ++ 
                                                                                                  " term: " ++ show t)
                                                                ty -> error ("wrong type on Get"
                                                                            ++ "\n\ttype: " ++ show ty 
                                                                            ++ "\n\tterm: " ++ show t)

typeof_ts t@(ADes (Project int) _ m2) ctx = case typeof_ts (AMsg m2) ctx of
                                                            TCat xs -> if length xs >= int then xs!!(int-1) else error ("index too large/lenght: " ++ show int ++ "/"++ show (length xs) ++ " term: " ++ show t)
                                                            ty -> error ("wrong type on Project"
                                                                        ++ "\n\ttype: " ++ show ty 
                                                                        ++ "\n\tterm: " ++ show t) 

typeof_ts (ADes GetSignH m1 m2) ctx = case typeof_ts (AMsg m1) ctx of
                                                        TPublicKey -> case typeof_ts (AMsg m2) ctx of
                                                                        TSign (TPrivateKey,t@(THash _)) -> t
                                                                        _ -> error (show m2 ++ " must be a Sign(Private,Hash(_)))") 
                                                        _ -> error (show m1 ++ " must be a PublicKey")
                                                      
-- verifiers

typeof_ts (AVer Vsign m1 m2) ctx = case typeof_ts (AMsg m1) ctx of
                                            TPublicKey -> case typeof_ts (AMsg m2) ctx of
                                                            TSign (TPrivateKey,_) -> TBoolean
                                                            _ -> error (show m2 ++ " must be a Sign(PrivateKey,t)") 
                                            _ -> error (show m1 ++ " must be a PublicKey")
                                                                                
typeof_ts (AVer Vcrypt m1 m2) ctx = case typeof_ts (AMsg m1) ctx of
                                                        TPrivateKey -> case typeof_ts (AMsg m2) ctx of
                                                                        TCrypt (TPublicKey,_) -> TBoolean
                                                                        _ -> error (show m2 ++ " must be a Crypt(PublicKey,_)") 
                                                        _ -> error (show m1 ++ " must be a PrivateKey")
typeof_ts (AVer Vscrypt m1 m2) ctx = let
                                            t1 = typeof_ts (AMsg m1) ctx
                                            t2 = typeof_ts (AMsg m2) ctx 
                                     in case t2 of
                                                TScrypt (t3,_) -> if compareTypes t1 t3 then TBoolean else error (show m2 ++ " must be a Scrypt(" ++ show t1 ++ ",_)") 
                                                _ -> error (show m2 ++ " must be a Scrypt(" ++ show t1 ++ ",_)")

typeof_ts t@(AVer Vformat (Atom ('f':'r':'m':id)) m2) ctx = let 
                                                                t1 = typeof_ts (AMsg (Atom id)) ctx        
                                                                t2 = typeof_ts (AMsg m2) ctx 
                                                            in case t1 of
                                                              TFormat _ _  -> if compareTypes t1 t2 then TBoolean else error ("wrong format expected: " ++ show t1 ++ "\nfound: " ++ show t2)
                                                              ty -> error ("wrong type on VFormat"
                                                                        ++ "\n\ttype: " ++ show ty 
                                                                        ++ "\n\tterm: " ++ show t)
typeof_ts (AVer Check m1 m2) ctx = let 
                                            t1 = typeof_ts (AMsg m1) ctx
                                            t2 = typeof_ts (AMsg m2) ctx 
                                   in if compareTypes t1 t2 then TBoolean else error (show m1 ++ " and " ++ show m2 ++ "must have the same type\n" ++ 
                                                                          show t1 ++ "/=" ++ show t2)

typeof_ts (AVer (VSignH msg) m1 m2) ctx = case typeof_ts (AMsg m1) ctx of --
                                            TPublicKey -> case typeof_ts (AMsg m2) ctx of
                                                            TSign (TPrivateKey,THash t) -> if compareTypes t (typeof_ts (AMsg msg) ctx) then TBoolean else error (show msg ++ " must be of type: " ++ show t)
                                                            _ -> error (show m2 ++ " must be a Sign(PrivateKey,Hash(_))") 
                                            _ -> error (show m1 ++ " must be a PublicKey")

typeof_ts (AVer VFsignH m1 m2) ctx = case typeof_ts (AMsg m1) ctx of
                                            TPublicKey -> case typeof_ts (AMsg m2) ctx of
                                                            TSign (TPrivateKey,THash _) -> TBoolean
                                                            _ -> error (show m2 ++ " must be a Sign(PrivateKey,Hash(_))") 
                                            _ -> error (show m1 ++ " must be a PublicKey")

typeof_ts t _ = error ("type of " ++ show t ++ " is undefined")

--typeof_ts t@(AVer v m1 m2) ctx = TBoolean
--typeof_ts t@(ADes d m1 m2) ctx = error ("type of " ++ show t ++ " is undefined\n" 
--                                                       ++ "d: " ++ show d  ++ " unhandled")

